/*-
 * Public Domain 2014-present MongoDB, Inc.
 * Public Domain 2008-2014 WiredTiger, Inc.
 *
 * This is free and unencumbered software released into the public domain.
 *
 * Anyone is free to copy, modify, publish, use, compile, sell, or
 * distribute this software, either in source code form or as a compiled
 * binary, for any purpose, commercial or non-commercial, and by any
 * means.
 *
 * In jurisdictions that recognize copyright laws, the author or authors
 * of this software dedicate any and all copyright interest in the
 * software to the public domain. We make this dedication for the benefit
 * of the public at large and to the detriment of our heirs and
 * successors. We intend this dedication to be an overt act of
 * relinquishment in perpetuity of all present and future rights to this
 * software under copyright law.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY CLAIM, DAMAGES OR
 * OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
 * OTHER DEALINGS IN THE SOFTWARE.
 */

#include <wiredtiger_config.h>
#include <inttypes.h>
#include <stddef.h>
#include <string.h>

#if defined(__linux__) && !defined(HAVE_NO_CRC32_HARDWARE)
#include <asm/hwcap.h>
#include <sys/auxv.h>

#ifndef __GNUC__
#define __asm__ asm
#endif

#define CRC32CX(crc, value) \
    __asm__("crc32cx %w[c], %w[c], %x[v]" : [c] "+r"(*&crc) : [v] "r"(+value))
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wunused-macros"
#define CRC32CW(crc, value) \
    __asm__("crc32cw %w[c], %w[c], %w[v]" : [c] "+r"(*&crc) : [v] "r"(+value))
#define CRC32CH(crc, value) \
    __asm__("crc32ch %w[c], %w[c], %w[v]" : [c] "+r"(*&crc) : [v] "r"(+value))
#pragma GCC diagnostic pop
#define CRC32CB(crc, value) \
    __asm__("crc32cb %w[c], %w[c], %w[v]" : [c] "+r"(*&crc) : [v] "r"(+value))

/*
 * __wt_checksum_with_seed_hw --
 *     Return a checksum for a chunk of memory, computed in hardware using 8 byte steps. Start with
 *     the given seed.
 */
static uint32_t
__wt_checksum_with_seed_hw(uint32_t seed, const void *chunk, size_t len)
{
    uint32_t crc;
    uint64_t uint64_v;
    uint32_t uint32_v;
    uint16_t uint16_v;
    const uint8_t *p;

    crc = ~seed;

    p = (const uint8_t *)chunk;
    /*
     * Checksum in 16B chunks. Do two loads at a time as the looping logic generated by the compiler
     * is faster.
     */
    for (; len >= 2 * sizeof(uint64_t); p += 2 * sizeof(uint64_t), len -= 2 * sizeof(uint64_t)) {
        memcpy(&uint64_v, p, sizeof(uint64_t));
        CRC32CX(crc, uint64_v);
        memcpy(&uint64_v, p + sizeof(uint64_t), sizeof(uint64_t));
        CRC32CX(crc, uint64_v);
    }

    if (len & sizeof(uint64_t)) {
        memcpy(&uint64_v, p, sizeof(uint64_t));
        CRC32CX(crc, uint64_v);
        p += sizeof(uint64_t);
    }

    if (len & sizeof(uint32_t)) {
        memcpy(&uint32_v, p, sizeof(uint32_t));
        CRC32CW(crc, uint32_v);
        p += sizeof(uint32_t);
    }

    if (len & sizeof(uint16_t)) {
        memcpy(&uint16_v, p, sizeof(uint16_t));
        CRC32CH(crc, uint16_v);
        p += sizeof(uint16_t);
    }

    if (len & sizeof(uint8_t))
        CRC32CB(crc, *p);

    return (~crc);
}

/*
 * __wt_checksum_hw --
 *     Return a checksum for a chunk of memory, computed in hardware using 8 byte steps.
 */
static uint32_t
__wt_checksum_hw(const void *chunk, size_t len)
{
    return (__wt_checksum_with_seed_hw(0, chunk, len));
}
#endif

extern uint32_t __wt_checksum_sw(const void *chunk, size_t len);
extern uint32_t __wt_checksum_with_seed_sw(uint32_t, const void *chunk, size_t len);
#if defined(__GNUC__)
extern uint32_t (*wiredtiger_crc32c_func(void))(const void *, size_t)
  __attribute__((visibility("default")));
extern uint32_t (*wiredtiger_crc32c_with_seed_func(void))(uint32_t seed, const void *, size_t)
  __attribute__((visibility("default")));
#else
extern uint32_t (*wiredtiger_crc32c_func(void))(const void *, size_t);
extern uint32_t (*wiredtiger_crc32c_with_seed_func(void))(uint32_t seed, const void *, size_t);
#endif

/*
 * wiredtiger_crc32c_func --
 *     WiredTiger: detect CRC hardware and return the checksum function.
 */
uint32_t (*wiredtiger_crc32c_func(void))(const void *, size_t)
{
    static uint32_t (*crc32c_func)(const void *, size_t);
#if defined(__linux__) && !defined(HAVE_NO_CRC32_HARDWARE)
    unsigned long caps;
#endif

    /*
     * This function calls slow hardware functions; if the application doesn't realize that, they
     * may call it repeatedly rather than caching the result.
     */
    if (crc32c_func != NULL)
        return (crc32c_func);

#if defined(__linux__) && !defined(HAVE_NO_CRC32_HARDWARE)
    caps = getauxval(AT_HWCAP);
    if (caps & HWCAP_CRC32)
        return (crc32c_func = __wt_checksum_hw);
    return (crc32c_func = __wt_checksum_sw);
#else
    return (crc32c_func = __wt_checksum_sw);
#endif
}

/*
 * wiredtiger_crc32c_with_seed_func --
 *     WiredTiger: detect CRC hardware and return the checksum function that accepts a starting
 *     seed.
 */
uint32_t (*wiredtiger_crc32c_with_seed_func(void))(uint32_t, const void *, size_t)
{
    static uint32_t (*crc32c_func)(uint32_t, const void *, size_t);
#if defined(__linux__) && !defined(HAVE_NO_CRC32_HARDWARE)
    unsigned long caps;
#endif

    /*
     * This function calls slow hardware functions; if the application doesn't realize that, they
     * may call it repeatedly rather than caching the result.
     */
    if (crc32c_func != NULL)
        return (crc32c_func);

#if defined(__linux__) && !defined(HAVE_NO_CRC32_HARDWARE)
    caps = getauxval(AT_HWCAP);
    if (caps & HWCAP_CRC32)
        return (crc32c_func = __wt_checksum_with_seed_hw);
    return (crc32c_func = __wt_checksum_with_seed_sw);
#else
    return (crc32c_func = __wt_checksum_with_seed_sw);
#endif
}
